#include "include.h"

static post *FindPost (unsigned int key, unsigned int rkey)
{
	post *p = (post*) Globals.Posts.intFind (key);
	return (p && p->rkey == rkey) ? p : NULL;
}

static cryptotype goodCrypto (char *s)
{
	cryptotype c = ahtoi (s);
	return (c % 2 == 0 || c == 1) ? 0 : c;
}

/* Download */
static void do_download (i_argstream &I)
{
	char tmp [MAX_MSGID];
	unsigned int key, rkey;
	post *p;
	o_argstream O;
	bool b;

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (tmp, MAX_MSGID) != GETARG_OK
	|| !parseID (tmp, &key, &rkey)) return;

	sem_wait (&Globals.DatabaseLock);  //################################
	if ((b = (p = FindPost (key, rkey))))
		p->buildDOWNLOAD (O);
	sem_post (&Globals.DatabaseLock);  //////////////////////////////////

	if (!b) O.putarg (Globals.PMSGtext [NOT_AVAILABLE]);

	O.writeout (I.fd);
}


/* Get */
static void do_get (i_argstream &I)
{
	char tmp [MAX_MSGID], oas [70], dateb [MAX_DATE];
	unsigned int key, rkey, nr = 0;
	time_t t = 0;
	int sz;
	post *p;
	o_argstream O (oas, sizeof oas);

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (tmp, MAX_MSGID) != GETARG_OK
	|| !parseID (tmp, &key, &rkey)) return;

	sem_wait (&Globals.DatabaseLock);  //################################
	if ((p = FindPost (key, rkey))) {
		nr = p->replies.cnt;
		t = p->date;
	}
	sem_post (&Globals.DatabaseLock);  //////////////////////////////////

	if (p) {
		O.putarg (Globals.PMSGtext [PNET_OK]);
		O.putarg ((unsigned int) posttextsize (key));
		O.putzero ();
		O.putarg (nr);
		O.putarg (network_date (t, dateb));
		O.writeout (I.fd);

		if (I.getarg (tmp, MAX_MSGID) == GETARG_OK
		&& PMSGtype (tmp) == GET)
			if ((sz = posttextsize (key)) > 0) {
				O.putarg (Globals.PMSGtext [PNET_OK]);
				O.writeout (I.fd);
				send_posttext (key, I.fd);
			}
			else goto na;

		return;
	}

na:
	O.putarg (Globals.PMSGtext [NOT_AVAILABLE]);
	O.writeout (I.fd);
}


/* Give reply */
void do_somebody_wants_my_reply (i_argstream &I)
{
	char tmp [MAX_MSGID], title [MAX_TITLE], cb [16], cX [16];
	char oas [MAX_PRL + 20 + MAX_TITLE + MAX_CRC + Globals.fieldOverhead];
	cryptotype KEY, Cy, Cb, CX;
	unsigned int key, rkey;
	post *p;
	o_argstream O (oas, sizeof oas);
	bool B = false;

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (tmp, MAX_MSGID) != GETARG_OK
	|| !parseID (tmp, &key, &rkey)
	|| I.getarg (cb, 15) != GETARG_OK || !(Cb = goodCrypto (cb))
	|| I.getarg (cX, 15) != GETARG_OK || !(CX = goodCrypto (cX)))
		return;

	sem_wait (&Globals.DatabaseLock);  //################################
	if ((p = FindPost (key, rkey)))
		if ((B = p->secret == SECRET_UNPOSTED)) {
			strcpy (tmp, MsgIDofPRL (p->parent));
			strcpy (title, p->title);
		}

	if (!B) {
		O.putarg (Globals.PMSGtext [NOT_AVAILABLE]);
		O.writeout (I.fd);
		goto out;
	}

	O.putarg (Globals.PMSGtext [PNET_OK]);
	O.putarg (tmp);
	O.putarg (KeyGeneration (Cb, CX, &KEY, &Cy));
	O.putarg (title);
	O.putarg (postcrc (key));
	O.writeout (I.fd);

	if (I.getarg (tmp, MAX_MSGID) != GETARG_OK) goto out;

	if (PMSGtype (tmp) == REDONE) {
		if (I.getarg (cX, 15) != GETARG_OK || !(CX = goodCrypto (cX)))
			goto out;
		KEY = xyc (CX, Cy);
	} else if (PMSGtype (tmp) != PNET_OK) goto out;

	if ((p = FindPost (key, rkey)) && p->secret == SECRET_UNPOSTED)
		p->secret = KEY;

	Rposted (key, KEY);
out:
	sem_post (&Globals.DatabaseLock);  ///////////////////////////////////
}


/* Reply notification, Go get it! */
static void do_come_get_reply_notification (i_argstream &I)
{
	char msg [MAX_MSGID], cpoint [MAX_PRL], author [MAX_AUTHOR], tmp [5];
	char pmsg [MAX_MSGID], cY [16], title [MAX_TITLE], crc [MAX_CRC];
	unsigned int key, rkey;
	cryptotype Cb, Cx, CY;
	char oas1 [10], oas2 [2 * MAX_MAX];
	o_argstream O (oas1, sizeof oas1), O2 (oas2, sizeof oas2);
	i_argstream I2;
	int version;
	post *p;
	bool redone = false, posted = false;

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (msg, MAX_MSGID) != GETARG_OK
	|| !parseID (msg, &key, &rkey)
	|| I.getarg (cpoint, MAX_PRL) != GETARG_OK) return;

	O.putarg (Globals.PMSGtext [PNET_OK]);
	O.writeout (I.fd);

	// now we connect, and the other will execute
	// symetrically do_somebody_wants_my_reply ()
	if (!Banned (cpoint) && contact (cpoint, I2, author, version)) {
		O2.putarg (Globals.PMSGtext [YOU_HAVE_REPLY]);
		O2.putarg (msg);
		O2.putarg (Cb = (cryptorand () % 100 + 2) | 1);
		O2.putarg (xyc (Cb, Cx = cryptorand ()));
		O2.writeout (I2.fd);

		if (I2.getarg (tmp, 10) != GETARG_OK
		|| PMSGtype (tmp) != PNET_OK
		|| I2.getarg (pmsg, MAX_MSGID) != GETARG_OK
		|| !parseID (pmsg, &key, &rkey)
		|| I2.getarg (cY, 10) != GETARG_OK || !(CY = goodCrypto (cY))
		|| I2.getarg (title, MAX_TITLE) != GETARG_OK
		|| I2.getarg (crc, MAX_CRC) != GETARG_OK)
			goto no_reply;


		// probable other fields ignored
		while (I2.peek ()) I2.getarg (tmp, MAX_MAX);

		sem_wait (&Globals.DatabaseLock);  //########################
		if ((p = FindPost (key, rkey))) {
			if ((redone = p->checksecret (xyc (CY, Cx))))
				do Cx = cryptorand ();
				while (p->checksecret (xyc (CY, Cx)));
			posted = newreply (p, cpoint, msg, title, author,
				 	ahtoi (crc), xyc (CY, Cx)); 
		}
		sem_post (&Globals.DatabaseLock);  //////////////////////////

		if (!p || !posted)
			O2.putarg (Globals.PMSGtext [NOT_AVAILABLE]);
		else if (!redone)
			O2.putarg (Globals.PMSGtext [PNET_OK]);
		else {
			O2.putarg (Globals.PMSGtext [REDONE]);
			O2.putarg (xyc (Cb, Cx));
		}
		O2.writeout (I2.fd);

no_reply:
		close (I2.fd);
	}
}


// ####### Attention. 
// ####### Very complicated. The four functions below
// ####### have a pervErted symmetry.

/* Contact point change notification (of reply) */
static void do_replier_change_notify (i_argstream &I)
{
	// lets explain this. This function called means:
	// somebody replied to our post (thus his reply
	// can be found). This person changed contact.
	// Check that he knows the secret, contact the
	// new contact-point, ask for verification.
	// per RFC, we are talking to John Bop.
	char msg [MAX_MSGID], sec [16], np [MAX_PRL];
	char author [MAX_AUTHOR], cY [16];
	char to_queuer_o [MAX_PRL], *g = NULL;
	unsigned int key, rkey;
	cryptotype CY, Cb, Cx, Secret = 0;
	char oas1 [30], oas2 [30];
	o_argstream O (oas1, sizeof oas1), O2 (oas2, sizeof oas2);
	i_argstream I2;
	int version;
	post *p;
	bool redone;

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (msg, MAX_MSGID) != GETARG_OK
	|| !parseID (msg, &key, &rkey)
	|| I.getarg (sec, 15) != GETARG_OK
	|| I.getarg (np, MAX_PRL) != GETARG_OK) return;

	sem_wait (&Globals.DatabaseLock);  //################################
	if ((redone = (p = FindPost (key, rkey))
		&& (g = p->checksecret (Secret = (cryptotype) ahtoi (sec))))) {
			CPofPRL (g, to_queuer_o);
			strcpy (msg, MsgIDofPRL (g));
		}
	sem_post (&Globals.DatabaseLock);  //////////////////////////////////

	if (!redone) {
		O.putarg (Globals.PMSGtext [NOT_AVAILABLE]);
		O.writeout (I.fd);
		return;
	}

	O.putarg (Globals.PMSGtext [PNET_OK]);
	O.writeout (I.fd);

	if (contact (np, I2, author, version)) {
		O2.putarg (Globals.PMSGtext [VERIFY_R]);
		O2.putarg (msg);
		O2.putarg (Secret);
		O2.putarg (Cb = (cryptorand () % 100 + 2) | 1);
		O2.putarg (xyc (Cb, Cx = cryptorand ()));
		O2.writeout (I2.fd);

		if (I2.getarg (msg, 5) == GETARG_OK
		&& PMSGtype (msg) == PNET_OK
		&& I2.getarg (cY, 15) == GETARG_OK && (CY = goodCrypto (cY))) {

			sem_wait (&Globals.DatabaseLock);  //#################
			if ((p = FindPost (key, rkey))) {
				if ((redone = p->checksecret (xyc (CY, Cx))))
					do Cx = cryptorand ();
					while (p->checksecret (xyc (CY, Cx)));
				p->pcrchange (Secret, np, xyc (CY, Cx));
				Rupdate_r (key, Secret, xyc (CY, Cx));
			}
			sem_post (&Globals.DatabaseLock);  ///////////////////

			if (!redone) O2.putarg (Globals.PMSGtext [PNET_OK]);
			else {
				O2.putarg (Globals.PMSGtext [REDONE]);
				O2.putarg (xyc (Cb, Cx));
			}
			O2.writeout (I2.fd);


			close (I2.fd);
			QAdapt (to_queuer_o, np);
			return;
		}
		// else if OK was sent but we didn't receive it
		// we are in trouble. Secret token generation lost
		// and can't be created again, except with
		// GURU arrangements
		close (I2.fd);
	}
}

/* Verify contact point change (with parent) */
static void do_verify_change_with_parent (i_argstream &I)
{
	// per RFC, this is John Bop in the callback (hi!)
	char msg [MAX_MSGID], sec [16], cb [16], cX [16], tmp [20];
	unsigned int key, rkey;
	cryptotype KEY, Cy, Cb, CX;
	post *p;
	bool B = false;
	char oas [30];
	o_argstream O (oas, sizeof oas);

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (msg, MAX_MSGID) != GETARG_OK
	|| !parseID (msg, &key, &rkey)
	|| I.getarg (sec, 15) != GETARG_OK
	|| I.getarg (cb, 15) != GETARG_OK || !(Cb = goodCrypto (cb))
	|| I.getarg (cX, 15) != GETARG_OK || !(CX = goodCrypto (cX))) return;

	sem_wait (&Globals.DatabaseLock);  //################################
	if ((p = FindPost (key, rkey)))
		if (B = (p->zombie) && p->secret == ahtoi (sec))
			p->zombie = false;

	if (B) {
		O.putarg (Globals.PMSGtext [PNET_OK]);
		O.putarg (KeyGeneration (Cb, CX, &KEY, &Cy));
		O.writeout (I.fd);


		if (I.getarg (tmp, MAX_MSGID) != GETARG_OK) goto out;
		if (PMSGtype (tmp) == REDONE) {
			if (I.getarg (cX, 15) != GETARG_OK
			|| !(CX = goodCrypto (cX))) goto out;
			KEY = xyc (CX, Cy);
		} else if (PMSGtype (tmp) != PNET_OK) goto out;


		Rupdate_p (p->Key, KEY);
		p->secret = KEY;
	} else {
		O.putarg (Globals.PMSGtext [NOT_AVAILABLE]);
		O.writeout (I.fd);
	}
out:
	sem_post (&Globals.DatabaseLock);  //////////////////////////
}

/* Verify contact point change (with reply) */
static void do_verify_change_with_reply (i_argstream &I)
{
	// lets explain this. This function called means:
	// somebody who replied to me wants to verify that
	// my contact point has changed. He supplies,
	// the message ID, the replyPRL, the secret.
	char msg [MAX_MSGID], sec [16], cb [16], cX [16];
	unsigned int key, rkey;
	cryptotype KEY, CY=0, Cy, Secret, Cb, CX;
	char oas [30];
	o_argstream O (oas, sizeof oas);
	post *p;
	bool B;

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (msg, MAX_MSGID) != GETARG_OK
	|| !parseID (msg, &key, &rkey)
	|| I.getarg (sec, 15) != GETARG_OK
	|| I.getarg (cb, 15) != GETARG_OK || !(Cb = goodCrypto (cb))
	|| I.getarg (cX, 15) != GETARG_OK || !(CX = goodCrypto (cX))) return;

	Secret = (cryptotype) ahtoi (sec);

	sem_wait (&Globals.DatabaseLock);  //################################
	if (B = (p = FindPost (key, rkey)) && p->update_ready (Secret)) {
		do CY = KeyGeneration (Cb, CX, &KEY, &Cy);
		while (p->checksecret (KEY));

		p->updated (Secret, KEY);
		Rupdate_r (p->Key, Secret, KEY);
	}
	sem_post (&Globals.DatabaseLock);  //////////////////////////////////

	if (B) {
		O.putarg (Globals.PMSGtext [PNET_OK]);
		O.putarg (CY);
	} else O.putarg (Globals.PMSGtext [NOT_AVAILABLE]);

	O.writeout (I.fd);
}

/* Contact point change notification (of reply) */
static void do_parent_change_notify (i_argstream &I)
{
	// lets explain this. This function called means:
	// the parent of my post belongs in a person
	// who changes his contact point. He supplies my
	// message id (i know its parent) and the secret
	// Then i go verify that with the callback.
	char msg [MAX_MSGID], sec [16], np [MAX_PRL];
	char author [MAX_AUTHOR], cY [16], pmsg [MAX_MSGID];
	char to_queuer_o [MAX_PRL];
	unsigned int key, rkey;
	cryptotype Secret, Cb, Cx, CY;
	char oas1 [30], oas2 [30];
	o_argstream O (oas1, sizeof oas1), O2 (oas2, sizeof oas2);
	i_argstream I2;
	int version;
	post *p;

DPRINTF (DB_SLOT4, "+++Enter %s\n", __FUNCTION__);
	if (I.getarg (msg, MAX_MSGID) != GETARG_OK
	|| !parseID (msg, &key, &rkey)
	|| I.getarg (sec, 15) != GETARG_OK
	|| I.getarg (np, MAX_PRL) != GETARG_OK) return;

	sem_wait (&Globals.DatabaseLock);  //################################
	p = FindPost (key, rkey);
	Secret = p->secret;
	strcpy (pmsg, MsgIDofPRL (p->parent));
	CPofPRL (p->parent, to_queuer_o);
	sem_post (&Globals.DatabaseLock);  //////////////////////////////////

	if (!p || Secret != ahtoi (sec)) {
		O.putarg (Globals.PMSGtext [NOT_AVAILABLE]);
		O.writeout (I.fd);
		return;
	}

	O.putarg (Globals.PMSGtext [PNET_OK]);
	O.writeout (I.fd);

	if (contact (np, I2, author, version)) {

		O2.putarg (Globals.PMSGtext [VERIFY_P]);
		O2.putarg (pmsg);
		O2.putarg (Secret);
		O2.putarg (Cb = (cryptorand () % 100 + 2) | 1);
		O2.putarg (xyc (Cb, Cx = cryptorand ()));
		O2.writeout (I2.fd);

		if (I2.getarg (msg, 5) == GETARG_OK
		&& PMSGtype (msg) == PNET_OK
		&& I2.getarg (cY, 15) == GETARG_OK && (CY = goodCrypto (cY))) {

			sem_wait (&Globals.DatabaseLock);  //################
			if ((p = FindPost (key, rkey)))
				p->pcpchange (Secret, np, xyc (CY, Cx));
			Rupdate_p (key, xyc (CY, Cx));
			sem_post (&Globals.DatabaseLock);  //////////////////

			close (I2.fd);
			QAdapt (to_queuer_o, np);
			return;
		}

		close (I2.fd);
	}
}

// #################
// #################

static void *client_th (void *vfd)
{
	int fd = *(int*)vfd;
	i_argstream I (fd);
	char mtype [5];

	delete (int*) vfd;
	++Globals.nClients;

DPRINTF (DB_SLOT4, "++++Enter %s (fd=%i)\n", __FUNCTION__, fd);
	write (fd, Globals.Greeting.Text, Globals.Greeting.size);

	if (I.getarg (mtype, 5) == GETARG_OK)
	switch (PMSGtype (mtype)) {
	case DOWNLOAD:
		do_download (I);
		break;
	case GET:
		do_get (I);
		break;
	case HAVE_REPLY:
		do_come_get_reply_notification (I);
		break;
	case YOU_HAVE_REPLY:
		do_somebody_wants_my_reply (I);
		break;
	case POINT_CHANGE_REPLY:
		do_replier_change_notify (I);
		break;
	case VERIFY_R:
		do_verify_change_with_parent (I);
		break;
	case POINT_CHANGE_PARENT:
		do_parent_change_notify (I);
		break;
	case VERIFY_P:
		do_verify_change_with_reply (I);
		break;
	case PING:
	default:
		break;
	}

	close (fd);
	--Globals.nClients;

	return NULL;
}

#define MAX_CLIENTS 2000

int do_client (int fd)
{
	if (Globals.nClients < MAX_CLIENTS) {
		pthread_t t;

		pthread_create (&t, NULL, client_th, new int (fd));
		pthread_detach (t);
	} else close (fd);

	return 0;
}

/* HCC RESOLVE:
pthread_create()
	client_th (Pv)
END */
